/*
** Dynamic Forking Proxy server for content switch project
**			ganesh kumar godavari (gkgodava@archie.uccs.edu)
*/

/* general libraries */
#include <setjmp.h>

#include "msock.h"

#define MIN(a,b) 		(a>b) ? b : a

char *log_file;	


union semun {
  int val;
  struct semid_ds *buf;
  ushort *array;
};

int semID;           			/* ID of the semaphore set.       	*/
union semun sem_val;  			/* semaphore value, for semctl(). 	*/

static int long_jumped=0,g_sock_fd;
static sigjmp_buf jmpbuf;
static volatile sig_atomic_t canjump;
static FILE *log_fp = (FILE *) NULL;

static void serveRequest(int fd,struct front_server *remote_ip);

void fatalError(char *format,...)
{
    va_list  args;

    va_start(args,format);
    (void) fprintf (stderr,"Error: ");
    vfprintf(stderr,format,args);
    va_end(args);
    exit(-1);
}

/*
** write all the junk!!  which u think may be useful to Administrator
**
*/
void logPrint(FILE * log_fp,char * format,...)
{
    va_list  args;
    time_t   now;

 
   char   *time_string;

    if(log_fp == NULL) {
	(void) fprintf(stderr,"unable to identify how the log file %s closed \n",LOG_FILE);
	return;
    }
	
    now=time(NULL);

    time_string=ctime(&now);
    time_string[strlen(time_string)-1]='\0';

    va_start(args,format);

    (void) fprintf(log_fp,"[%s]",time_string);
    (void) vfprintf(log_fp,format,args);

    va_end(args);
    fflush(log_fp);
    return;

}

static void sigAlarm(int sigtype)
{
    if (canjump == 0)
    {
        return;
    }
    canjump=1;
    long_jumped=1;
    siglongjmp(jmpbuf,1);
}

static u_long pow(int a, int b)
{
u_long temp=0;

if (b == 0) return 1;

if (b == 1) return a;

for ( ;b>1; b--)
  temp = temp + (a*a);

return temp;
}

/*
** function converts hexadecimal string to a decimal integer.
*/
static u_long Hex(char *s)
{
int pos=0;
u_long value=0;
int len = strlen(s)-2; /* skipping the \r\n characters here*/

for(pos = 0, s += (len-1),value=0 ;pos<len ;s--,pos++)
{
  switch(*s)
  {
  case 'a':
  case 'A':
           value = value + (pow(16,pos) * 10);
           break;


  case 'b':
  case 'B':
           value = value + (pow(16,pos) * 11);
           break;

  case 'c':
  case 'C':
           value = value + (pow(16,pos) * 12);
           break;


  case 'd':
  case 'D':
           value = value + (pow(16,pos) * 13);
           break;

  case 'e':
  case 'E':
           value = value + (pow(16,pos) * 14);
           break;

  case 'f':
  case 'F':
           value = value + (pow(16,pos) * 15);
           break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
           value = value + (pow(16,pos) * (*s - '0'));
           break;
 default:
         return value;
         break;
  }
 }
 return value;
}

void make_absolute_path(void)
{
 if ((log_file   = (char *) malloc (sizeof (SERVER_ROOT) + sizeof (LOG_FILE) + 1)) == NULL) 
  {
   (void) fprintf(stderr," unable to allocate memory for converting relative paths to absolute paths");
   exit(1);
  }
(void) sprintf(log_file,"%s/%s",SERVER_ROOT,LOG_FILE);
 return;
}

int main(void)
{
    char szbuf[BUFSIZ];
    int rc=0,sock_fd;

    struct front_server szclient;

    (void)make_absolute_path();     /* convert relative to absolute path */

    /* open the log file */
    if((log_fp=fopen(log_file,"a+"))== NULL) 
	(void) fprintf(stderr, "unable to open the log file %s", log_file);

    /* create a semaphore */
    semID = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600);
    if (semID == -1) {
        (void) fprintf(stderr,"Error: unable to create a semaphore\n");
        exit(1);
    }

    /* intialize the semaphore to '1'. */
    sem_val.val = 1;
    rc = semctl(semID, 0, SETVAL, sem_val);
    if (rc == -1)
    {
        (void) fprintf(stderr,"Error: unable to initialize the semaphore");
        exit(1);
    } 

    /*
    ** most cool Unix will take care of Zombies if we ignore SIGCHLD
    */
    signal(SIGCHLD,SIG_IGN);
    signal(SIGPIPE,SIG_IGN);
    signal(SIGALRM,sigAlarm);

    noCoredump(); 
    daemonize(); 

    /*
    ** open the server socket
    */
    sock_fd=serverSocket((u_short) SERVER_PORT ,1000);
    if (sock_fd == -1)
    {
        (void) fprintf(stderr,"Failed to create server\n");
        exit(1);
    }
    g_sock_fd=sock_fd;

    /*
    ** get name, IP adress, and port number of the client
    */
    getPeerInfo(sock_fd,&szclient);
    open_critical_lock();

    alarm(CLIENT_TIMEOUT);

    sigsetjmp(jmpbuf,1);
    if (long_jumped == 1)
    {
        /*
        ** back from a long jump, alarm timer expired
        */
        long_jumped=0;
        canjump=1;
        goto ExitProcessing;
    }

    serveRequest(sock_fd,&szclient);


ExitProcessing:
    close(sock_fd);
    exit(1);

}

/*
** here i am fixing the method and protocol name field to 10 bytes. i feel allocating  10 bytes is waste of "resource"
*/
static int parseRequest(char *req,char *method,char *url,char *proto,char *args,int len)
{
    char
        *chp,
        *m,
        *u,
        *p;

    if (req == NULL || *req == '\0')
        return (-1);

    /*
    ** Method
    */
    m=strtok(req,"\t ");
    if (m == NULL)
        return(-1);
    (void) strncpy(method,m,10);
    method[10]='\0';

    /*
    ** Url
    ** Url can't have any spaces
    */
    u=strtok(NULL,"\t\r ");
    if (u == NULL)
        return(-1);
    chp=strchr(u,'?');
    if (chp)
    {
        *chp++ = '\0';
        (void) strncpy(args,chp,len);
    }
 
    (void) strncpy(url,u,len);
    url[len]='\0';

    /*
    ** HTTP Protocol version
    */
    p=strtok(NULL,"\r ");
    if (p == NULL)
        return(-1);
    (void) strncpy(proto,p,10);
    proto[10]='\0';

    return (0); /* OK */
}


/*
** find a substring from a string
** case insensitive
*/
int stristr(char *haystack,char *needle)
{
    int
        i, j, k, l;

    for (i = 0; haystack[i]; i++)
    {
        for (j = 0, l = k = i; haystack[k] && needle[j] &&
            tolower(haystack[k]) == tolower(needle[j]); j++, k++)
                ;
        if (needle[j] == '\0')
            return (i);
    }
    return -1;
}


/*
** returns the formated text i.e replacing the text containing source with dest
** none of the arguments are modified
*/
char *replace(char *text, char *source, char *dest)
{
  char *nh,newtext[1024];
  int start=0,end=0;

  if((start=stristr(text,source)) == -1)
  {
    (void) fprintf(stderr," unable to located substring %s in the string %s\n",text,source);
  }
#if 0
 fprintf(stderr,"\nstart value is %d\n",start);
#endif

  if((nh = (char *) malloc(start+1))== NULL)
  {
   (void) fprintf(stderr, "unable to allocate memory for the header\n");
   return 0;
  }

  strncpy(nh,text,start);
  nh[start]='\0';

  end = strlen(source);

strcpy(newtext,nh);
#if 0
    (void) fprintf(stderr," new text %s\n",newtext);
#endif
strcat(newtext,dest);
#if 0
    (void) fprintf(stderr," new text %s\n",newtext);
#endif
strcat(newtext,(text+start+end));
strcpy(text,newtext);

/*
** when i am freeing the memory occupied by headers and nh it is causing SIGSEGV(Segmentation Fault) only for I.E not for netscape 
** damn these browers / damn my coding as one of these is wrong
**  free(header);
**  free(nh);
*/
 return text;
} 


void demoResponse(int fd, char *url)
{
 char body[1024];
 char content_length[64];
 sprintf(body," <!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">
                   <HTML><HEAD>
                   <TITLE>404 Not Found</TITLE>
                   </HEAD><BODY>
                   <H1>Not Found</H1>
                   The requested URL %s was not found on this server.<P>
                   <HR>
                   <ADDRESS>Apache/1.3.14 Server at %s  Port 80</ADDRESS>
                   </BODY></HTML>",url,SERVER_NAME);

 sprintf(content_length,"Content-Length: %d\r\n",strlen(body));

 sockPuts(fd,"HTTP/1.0 404 OK\r\n");
 sockPuts(fd,"MIME-version: 1.0\r\n");
 sockPuts(fd,content_length);
 sockPuts(fd,"Content-Type: text/html\r\n\r\n");
 sockPuts(fd,body);

 return;
}

/*
** serverRequest is currently clumsy -- fix it and make it readable(gkgodava@archie.uccs.edu)
*/
static void serveRequest(int fd,struct front_server  *remote_ip)
{
  char
     header[1024],
     chunked_body[1024],
     *val,
     szcontent_type[BUFSIZ],
     ruleheader[BUFSIZ],
     szmethod[10],
     szurl[BUFSIZ],
     szproto[10],
     szargs[BUFSIZ],
     body[IOBUFFSIZE],
     TargetServer[128],
     szbuf[BUFSIZ];


  int
     maxfd,
     nready,
     close_connection,
     chunked,
     content_length,
     server_sock_fd,
     chunk_size,
     status_code,
     data_read;

  float version;


  struct timeval  tv;
  Connection *conn_rec;
  u_long ret_addr;
  fd_set   readfds;

KEEPALIVE:
  /* if the client does not respond within the timeout limit, zap it  */
   FD_ZERO(&readfds);

   FD_SET(fd, &readfds);
   maxfd = fd + 1;

  content_length=0;
  data_read=0;
  *szcontent_type='\0';
  *szmethod='\0';
  *szurl='\0';
  *szproto='\0';
  *szargs='\0';
  *szbuf='\0';

   tv.tv_sec=CLIENT_TIMEOUT;
   tv.tv_usec=0;

    nready=select(maxfd,&readfds,NULL,NULL,&tv);
    if (nready == 0)
    {
        (void) fprintf(stderr," *** pid %d Timed out in %d seconds ***\n",getpid(),CLIENT_TIMEOUT);
        goto  GracefulExit;
    }

 /*
 ** get request
 */
 if(FD_ISSET(fd,&readfds))
  {
     alarm(0);
    /* check whether the packet sent after the connection established has no data -- if so reject it */
     if((sockGets(fd,szbuf,sizeof(szbuf)-1)) <= 0)
        goto  GracefulExit;
   }

  /*
  ** now we can allocate memory for the connection structue as we know it is not a junk request!!
  */
  if( ((conn_rec = (struct Connection *) malloc(sizeof(Connection))) == NULL) || 
       ((conn_rec->cli_ser = (struct front_server *) malloc(sizeof(struct front_server))) == NULL) || 
       ((conn_rec->req = (struct Request *) malloc(sizeof(Request)))  == NULL) ) 
   {
   	logPrint(log_fp,"Unable to create allocate memory for the connection structure !!\n");
        goto  GracefulExit;
   }

  memset(conn_rec,0,sizeof(conn_rec));

  conn_rec->req->header=NULL;

  strcpy(header, mystrdup(szbuf));

  if(parseRequest(szbuf,szmethod,szurl,szproto,szargs,BUFSIZ-1) != 0)
  {
	(void) fprintf(stderr," unable to process the request %s \n",szbuf);
        goto  GracefulExit;
  }

    logPrint(log_fp," request %s %s %s from %d \n",szmethod, szurl, szproto,remote_ip->addr);

    while(sockGets(fd,szbuf,sizeof(szbuf)-1) > 0)
        {

        strcpy(szargs,szbuf);

        val=strchr(szargs,':');
        if (val)
        {
            *val++ = '\0';
            if (strncasecmp("content-length",szbuf,strlen("content-length")) == 0)
                content_length=atoi(val);
            if (strncasecmp("content-type",szbuf,strlen("content-type")) == 0)
                (void) strcpy(szcontent_type,val);
            if (strncasecmp("Host",szbuf,strlen("Host")) == 0)
                (void) sprintf(szbuf,"%s %s\r\n","Host:",SERVER_NAME);

        }
        strcat(header,szbuf);
    }


   if((conn_rec->req->header = (char *) malloc(strlen(header)+1)) == NULL) 
    {
       (void) fprintf(stderr,"unable to allocate memory for conn->req->header\n");
       goto  GracefulExit;
    }
	
   (void) strcpy(conn_rec->req->header,header);

   if(content_length != 0)
   {
     if((conn_rec->req->body = (char *) malloc(content_length+1)) == NULL) 
       {
        (void) fprintf(stderr,"unable to allocate memory for conn->req->body\n");
        goto  GracefulExit;
       }
      sockRead(fd,conn_rec->req->body,content_length);
    }
     else
        conn_rec->req->body=NULL;

    /*
    ** store the connection information as i need to perform rule matching
    */
    memcpy(conn_rec->cli_ser,remote_ip,sizeof(struct front_server));
    strcpy(conn_rec->req->method,szmethod);
      
    if ((conn_rec->req->url = (char *) malloc(strlen(szurl)+1)) == NULL) 
     {
      (void) fprintf(stderr,"unable to allocate memory for conn->req->urlr\n");
      goto  GracefulExit;
     }
	
    (void) strcpy(conn_rec->req->url,szurl);

    if((conn_rec->req->content_type = (char *) malloc(strlen(szcontent_type)+1)) == NULL)
     {
       (void) fprintf(stderr,"unable to allocate memory for conn->req->content_type \r\n");
       goto  GracefulExit;
     }
	
    (void) strcpy(conn_rec->req->content_type,szcontent_type);

#if 0
    /*
    ** performing the rule matching
    */
    ret_addr = ruleMatching(conn_rec);

    (void) strcpy(TargetServer,getRealServerName(ret_addr));

#endif


    /*
    ** performing the rule matching
    ** this place is where i need to make a rule module call God Bless ME
    */

   if((server_sock_fd=connectTo(RULE_SERVER_NAME,RULE_SERVER_PORT)) == -1) {
     (void) fprintf(stderr,"Failed to connect to the rule module %s\n",SERVER_NAME);

   if((server_sock_fd=connectTo(DEFAULT_RULE_SERVER_NAME,DEFAULT_RULE_SERVER_PORT )) == -1) {
     (void) fprintf(stderr,"Failed to connect to the rule module %s\n",SERVER_NAME);
     goto  GracefulExit;
   }

   }

  (void) sprintf(ruleheader,"Port: %d\r\nClientAddr: %d\r\nMethod: %s\r\nUrl: %s\r\nContent-Type: %sContent-Length: %d\r\n\r\n", conn_rec->cli_ser->port, conn_rec->cli_ser->addr, conn_rec->req->method, conn_rec->req->url,conn_rec->req->content_type, content_length);

fprintf(stderr," rule header is %s",ruleheader);

  /* send the request to the Rule Module */
  sockWrite(server_sock_fd,ruleheader,strlen(ruleheader));
  if(strcmp(conn_rec->req->method, "POST") == 0)
    sockWrite(server_sock_fd,conn_rec->req->body,content_length);
(void) fprintf(stderr, "header details are %s", conn_rec->req->header);

 while(sockGets(server_sock_fd,szbuf,sizeof(szbuf)-1) > 0)
        {
        val=strchr(szbuf,':');
        if (val)
        {
            *val++ = '\0';
            if (strncasecmp("ServerAddress",szbuf,strlen("ServerAddress")) == 0)                ret_addr=atol(val);
            if (strncasecmp("ServerName",szbuf,strlen("ServerName")) == 0)
                (void) strcpy(TargetServer,val);
        }
    }



close(server_sock_fd) ;
TargetServer[strlen(TargetServer)-2] = '\0';

    /*
    ** done with the rule matching
    */



    if(szbuf == NULL)
    {
    (void) fprintf(stderr, "looks like server list has been modified\n");
    goto  GracefulExit;
    }
	
   conn_rec->req->header = mystrdup(replace(conn_rec->req->header,SERVER_NAME,TargetServer));

   if((server_sock_fd=connectTo(TargetServer,80 )) == -1)
   {
     (void) fprintf(stderr,"Failed to connect to %s\n",TargetServer);
     goto  GracefulExit;
   }

  /* send the request to the real server */
  sockPuts(server_sock_fd, conn_rec->req->header);
  sockPuts(server_sock_fd,"\r\n");

  if(strcmp(conn_rec->req->method, "POST") == 0)
     sockPuts(server_sock_fd,conn_rec->req->body);

  /* freeing & initializing the values*/
  content_length = 0;
  chunked = 0;
  close_connection = 0;
  status_code = 0;


 if((sockGets(server_sock_fd,szbuf,sizeof(szbuf)-1)) <= 0)
    goto  GracefulExit;

 /* check for the status code */
  sscanf(szbuf,"HTTP/%f %d \n",&version, &status_code);


  strcpy(header, mystrdup(szbuf));

  while(sockGets(server_sock_fd,szbuf,sizeof(szbuf)-1) > 0)
  {
   strcpy(szargs,szbuf);
   val=strchr(szbuf,':');

   if(val)
    {
     *val++ = '\0';
     if(strncasecmp("content-length",szbuf,strlen("content-length")) == 0)
       content_length=atoi(val);

     if(strncasecmp("Connection",szbuf,strlen("Connection")) == 0)
	if( strncasecmp(val," close",strlen(" close")) == 0)
          close_connection = 1;

    /*
    **  trying to make sure we follow rfc 2616 section 19.4.6 
    **  HTTP/1.1 introduces the Transfer-Encoding header field.
    **  Proxies/gateways MUST remove any transfer-coding prior to
    **  forwarding a message via a MIME-compliant protocol.
    */

     if(strncasecmp("Transfer-Encoding",szbuf,strlen("Transfer-Encoding")) == 0)
     {
      if( strncasecmp(val," chunked",strlen("chunked")) == 0)
       {
        chunked = 1;
        strcpy(szcontent_type,szargs);
       }
     }

    /*
    ** RFC 2616 sec 13.5.2
    ** A transparent proxy MUST NOT modify any of the following fields in a
    ** request or response, and it MUST NOT add any of these fields if not
    ** already present:
    **
    **  - Content-Location
    **
    **  - Content-MD5
    **
    **  - ETag
    **
    **  - Last-Modified
    **
    */
    /* i am violating this rule */
    if(strncasecmp("Location",szbuf,strlen("Location")) == 0)
    {
     strcpy(szargs,replace(szargs,TargetServer,SERVER_NAME));
#if 0
     strcpy(szargs,replace(szargs,"http","https"));
#endif
     (void) fprintf(stderr,"\nthe new location bar is %s\n",szargs);
    }
    }
  
   strcat(header,szargs);
   }

  if(status_code > 400 )
  {
   demoResponse(fd,szurl);
  }
  else if(chunked == 1)
  {
   chunk_size=0;
   *chunked_body='\0';
   content_length=0; /* chunked had more preference than content_length */

   if ( sockGets(server_sock_fd,body,IOBUFFSIZE-1) < 0 )

      fprintf(stderr, "unable to read chunked data");
   while((chunk_size=Hex(body)) != 0)
   {
    fprintf(stderr, "szbuf %s chunck size %d \n",body,chunk_size);
    
    content_length = content_length + chunk_size;
  
    while ((chunk_size > 0) && ((data_read=sockRead(server_sock_fd,body,MIN(IOBUFFSIZE-1,chunk_size))) > 0 ))
    {
     chunk_size = chunk_size - data_read;
     strcat(chunked_body,body);
    }
   
    if (sockGets(server_sock_fd,body,IOBUFFSIZE-1) < 0 )
       fprintf(stderr, "unable to read chunked data\n"); 
   }

   /* making changes to the header */
   sprintf(szbuf,"Content-Length: %d\r\n",content_length);
   strcpy(szargs, replace(header,szcontent_type,szbuf));

   strcpy(header,szargs);

  }

 if(status_code < 400)
 {
  /* sending headers to the client */ 
  strcat(header,"\r\n\0");
 
#if 0
 (void) fprintf(stderr," headers are %s\n",header);
#endif

  sockWrite(fd,header,strlen(header));

 /* sending the chunked data as body if the header contains Tansfer-Encoding: chunked\r\n*/
  if(chunked == 1)
  {
   sockWrite(fd,chunked_body,content_length);
   (void) fprintf(stderr," body is %s\n",chunked_body);
   content_length =0;
  }

  for(data_read=0; content_length > 0;data_read=0 )
  {
   if ((data_read = iread(server_sock_fd,body,IOBUFFSIZE-1)) < 0)
    {
      if (errno == ECONNRESET)
         (void) fprintf(stderr,"\nconnection was reset by the server\n");
      else
         (void) fprintf(stderr,"\nunable to read data from the socket\n");

      goto  GracefulExit;
     }
     content_length -= data_read;
     if(data_read != sockWrite(fd,body,data_read))
     {
       (void) fprintf(stderr,"pid %d unable to pump more data on content length\n",getpid()); 
       goto  GracefulExit; 
     }
  }

  (void) fprintf(stderr," pid %d data receive till now are  chunked %d close connection %d\n",getpid(),chunked,close_connection);
 }

  if(close_connection == 1)
    goto GracefulExit;
  else
  {
    sockPuts(fd,"\r\n");
    free(conn_rec->req->url);
    free(conn_rec->req->header);
    free(conn_rec->req->content_type);
    free(conn_rec->req->body);
    free(conn_rec->req);
#if 0
    free(conn_rec->cli_ser->name);
#endif
    free(conn_rec->cli_ser);
    (void) close(server_sock_fd);
    goto KEEPALIVE;
  }

GracefulExit:
 free(conn_rec->req->url);
 free(conn_rec->req->header);
 free(conn_rec->req->content_type);
 free(conn_rec->req->body);
 free(conn_rec->req);
#if 0
 free(conn_rec->cli_ser->name);
#endif
 free(conn_rec->cli_ser);
 (void) close(server_sock_fd);
 (void) close(fd); /* closing the connection*/
 (void) fprintf(stderr,"\n\n closing the ssl context bye bye %d ..\n",getpid());
  return;
}

